"""CLI helper functions."""

import argparse
import importlib
import os
from .__version__ import __version__

FILES_WITH_ARGS = [
    "r2x.plugins.pcm_defaults",
    "r2x.plugins.break_gens",
    "r2x.plugins.emission_cap",
    "r2x.plugins.hurdle_rate",
    "r2x.plugins.cambium",
    "r2x.parser.plexos",
    "r2x.parser.reeds",
    "r2x.exporter.plexos",
    "r2x.exporter.sienna",
]


class Flags(argparse.Action):
    """Class to enable feature flags on the code and CLI.

    This will save the feature flags in a `flags` argument in the NameSpace of
    argparse
    """

    def __call__(  # type: ignore
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        optional_str: str,
    ) -> None:
        """Save flags as arguments in Namespace."""
        setattr(namespace, self.dest, dict())

        for value in values:
            # split it into key and value
            key, value = value.split("=")
            # assign into dictionary
            getattr(namespace, self.dest)[key] = value


def get_additional_arguments(
    parser: argparse.ArgumentParser,
    folders: list[str] = FILES_WITH_ARGS,
) -> argparse.ArgumentParser:
    """Add plugin arguments to CLI.

    Args:
        parser: CLI parser object.
        plugin_list: List of plugins.
        folder: Folder that contains the plugin.
    """
    for package in folders:
        package_str = package.split(".")
        if not len(package_str) == 3:
            msg = (
                "Format of `FILES_WITH_ARGS` should be `r2x.package.script` but received {package_name}."
                "Modify `FILES_WITH_ARGS` to match the format."
            )
            raise NotImplementedError(msg)

        _, package_name, script_name = package_str
        try:
            package_script = importlib.import_module(f"{package}")
        except ImportError:
            continue
        if hasattr(package_script, "cli_arguments"):
            script_cli_group = parser.add_argument_group(f"{package_name.upper()}: {script_name}")
            package_script.cli_arguments(script_cli_group)

    return parser


def base_cli() -> argparse.ArgumentParser:
    """Create parser object for CLI."""
    parser = argparse.ArgumentParser(
        description="""Model translation framework""",
        add_help=True,
        prog="r2x",
    )
    subparsers = parser.add_subparsers(dest="command", help="Subcommands")
    init_command = subparsers.add_parser("init", help="Create an empty configuration file.")
    run_command = subparsers.add_parser("run", help="Run an R2X translation")

    init_command.add_argument(
        "-o",
        nargs="?",
        dest="path",
        default=os.getcwd(),
        help="Destination folder where the file will be copied. Defaults to current directory.",
    )

    group_run = run_command.add_argument_group("Options for running the code")
    group_run.add_argument("--inspect", action="store_true", help="Inspect resulting infrasys system.")
    group_run.add_argument("--upgrade", action="store_true", help="Run upgrader logic.")
    group = group_run.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-i",
        dest="run_folder",
        help="Location of the folder with the input model data",
    )
    group.add_argument(
        "--config",
        "--user-config",
        dest="user_config",
        help="User configuration",
    )
    group_cli = run_command.add_argument_group("Arguments used with a run folder")
    group_cli.add_argument(
        "--name",
        dest="name",
        help="Scenario name",
    )
    group_cli.add_argument(
        "--input-model",
        dest="input_model",
        choices=[
            "plexos",
            "reeds-US",
            "plexos",
            "infrasys",
        ],
        help="Input model to convert from",
    )
    group_cli.add_argument(
        "--output-model",
        dest="output_model",
        choices=["plexos", "sienna", "infrasys"],
        help="Output model to convert to",
    )
    group_cli.add_argument(
        "-o",
        nargs="?",
        dest="output_folder",
        help="Path to save the result translation",
    )
    group_cli.add_argument(
        "-y",
        "--year",
        type=int,
        nargs="+",
        dest="solve_year",
        help="Year to translate",
    )
    group_cli.add_argument(
        "--scenario",
        dest="scenario_name",
        help="Scenario to select from cases file",
    )
    group_cli.add_argument(
        "-ud",
        "--user-dict",
        dest="user_dict",
        help="User provided overrides to defaults",
    )
    group_cli.add_argument(
        "-p",
        "--plugins",
        dest="plugins",
        nargs="*",
        help="Plugins to be included in the translation",
    )
    group_cli.add_argument(
        "--save",
        action="store_true",
        help="Serialize infrasys system.",
    )
    run_command.add_argument("--pdb", action="store_true", dest="pdb", help="Run with debugger enabled.")
    run_command.add_argument("--flags", nargs="*", dest="feature_flags", action=Flags, help="Feature flags")
    parser.add_argument("--verbose", "-v", action="count", default=0, help="Run with additional verbosity")
    parser.add_argument("--version", "-V", action="version", version=f"R2X version: {__version__}")
    _ = get_additional_arguments(run_command)
    return parser



"""Set of helper functions for parsers."""
# ruff: noqa

from datetime import timedelta
from typing import Any
import polars as pl
import numpy as np
import cvxpy as cp

from infrasys.function_data import QuadraticFunctionData, PiecewiseLinearData, XYCoords


def field_filter(
    property_fields: dict[str, Any], eligible_fields: set[str]
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Filters a dictionary of property fields into valid and extra fields based on eligibility.

    Parameters
    ----------
    property_fields : dict
        Dictionary of property fields where keys are field names and values are field values.
    eligible_fields : set
        Set of field names that are considered valid (eligible).

    Returns
    -------
    tuple of dict
        A tuple of two dictionaries:
        - valid : dict
            Contains fields that are both in `property_fields` and `eligible_fields`, and are not `None`.
        - extra : dict
            Contains fields that are in `property_fields` but not in `eligible_fields`, and are not `None`.

    Examples
    --------
    >>> property_fields = {"field1": 10, "field2": None, "field3": "hello"}
    >>> eligible_fields = {"field1", "field2"}
    >>> valid, extra = field_filter(property_fields, eligible_fields)
    >>> valid
    {'field1': 10}
    >>> extra
    {'field3': 'hello'}
    """
    valid = {k: v for k, v in property_fields.items() if k in eligible_fields if v is not None}
    extra = {k: v for k, v in property_fields.items() if k not in eligible_fields if v is not None}

    return valid, extra


def prepare_ext_field(valid_fields: dict[str, Any], extra_fields: dict[str, Any]) -> dict[str, Any]:
    """Clean the extra fields by removing any time series data and adds the cleaned extra fields to `valid_fields`.

    Parameters
    ----------
    valid_fields : dict
        Dictionary containing valid fields.
    extra_fields : dict
        Dictionary containing extra fields that may include data types not needed.

    Returns
    -------
    dict
        Updated valid_fields with cleansed extra fields under the "ext" key.

    Examples
    --------
    >>> valid_fields = {"field1": 10, "field2": "hello"}
    >>> extra_fields = {"field3": [1, 2, 3], "field4": 42}
    >>> result = prepare_ext_field(valid_fields, extra_fields)
    >>> result
    {'field1': 10, 'field2': 'hello', 'ext': {'field4': 42}}
    """
    if extra_fields:
        # Filter to only include eligible data types
        eligible_datatypes = [str, int, float, bool]
        extra_fields = {k: v for k, v in extra_fields.items() if type(v) in eligible_datatypes}
        valid_fields["ext"] = extra_fields
    else:
        valid_fields["ext"] = {}
    return valid_fields


def handle_leap_year_adjustment(data_file: pl.DataFrame) -> pl.DataFrame:
    """Duplicate February 28th to February 29th for leap years.

    Parameters
    ----------
    data_file : pl.DataFrame
        DataFrame containing timeseries data.

    Returns
    -------
    pl.DataFrame
        DataFrame adjusted for leap years.

    Examples
    --------
    """
    if len(data_file) != 8760:
        raise ValueError("Data must contain 8760 rows for a non-leap year.")

    # Get the index positions for February 28th
    feb_28_start_index = 1392  # Start of February 28th (hour 0)
    feb_28_end_index = 1416  # End of February 28th (hour 24)

    # Slice February 28th data
    feb_28_data = data_file[feb_28_start_index:feb_28_end_index]

    # Create a new DataFrame with February 29th duplicated
    adjusted_data = pl.concat([data_file, feb_28_data])

    return adjusted_data


def fill_missing_timestamps(data_file: pl.DataFrame, hourly_time_index: pl.DataFrame) -> pl.DataFrame:
    """Add missing timestamps to data and forward fill nulls to complete a year.

    Parameters
    ----------
    data_file : pl.DataFrame
        DataFrame containing timeseries data.
    hourly_time_index : pl.DataFrame
        DataFrame containing the hourly time index for the study year.

    Returns
    -------
    pl.DataFrame
        DataFrame with missing timestamps filled.

    Raises
    ------
    ValueError
        If the required columns are missing from the data_file.

    Examples
    --------
    >>> import polars as pl
    >>> from datetime import datetime
    >>> df = pl.DataFrame(
    ...     {"year": [2020, 2020], "month": [1, 1], "day": [1, 1], "hour": [0, 1], "value": [1, 2]}
    ... )
    >>> hourly_time_index = pl.datetime_range(
    ...     datetime(2020, 1, 1), datetime(2020, 1, 2), interval="1h", eager=True, closed="left"
    ... ).to_frame("datetime")
    >>> fill_missing_timestamps(df, hourly_time_index)
    """
    column_set = set(data_file.columns)
    match column_set:
        # Case when "year", "month", "day", and "hour" are all present
        case s if s.issuperset({"year", "month", "day", "hour"}):
            data_file = data_file.with_columns(
                pl.datetime(pl.col("year"), pl.col("month"), pl.col("day"), pl.col("hour"))
            )
            upsample_data = hourly_time_index.join(data_file, on="datetime", how="left")
            return upsample_data.fill_null(strategy="forward")

        # Case when "year", "month", and "day" are present but "hour" is missing
        case s if s.issuperset({"year", "month", "day"}):
            data_file = data_file.with_columns(pl.datetime(pl.col("year"), pl.col("month"), pl.col("day")))
            upsample_data = hourly_time_index.join(data_file, on="datetime", how="left")
            return upsample_data.fill_null(strategy="forward")

        # Case when "day" is missing, but "year" and "month" are present
        case s if s.issuperset({"year", "month"}):
            data_file = data_file.with_columns(pl.datetime(pl.col("year"), pl.col("month"), 1))  # First day
            upsample_data = hourly_time_index.join(data_file, on="datetime", how="left")
            return upsample_data.fill_null(strategy="forward")

        case s if s.issuperset({"year", "datetime"}):
            upsample_data = hourly_time_index.join(data_file, on="datetime", how="left")
            return upsample_data.fill_null(strategy="forward")
        case _:
            raise ValueError("The data_file must have at least 'year' and 'month' columns.")


def resample_data_to_hourly(data_file: pl.DataFrame) -> pl.DataFrame:
    """Resample data to hourly frequency from minute data.

    Parameters
    ----------
    data_file : pl.DataFrame
        DataFrame containing timeseries data with minute intervals.

    Returns
    -------
    pl.DataFrame
        DataFrame resampled to hourly frequency.

    Examples
    --------
    >>> df = pl.DataFrame(
    ...     {
    ...         "year": [2020, 2020, 2020, 2020],
    ...         "month": [2, 2, 2, 2],
    ...         "day": [28, 28, 28, 28],
    ...         "hour": [0, 0, 1, 1],
    ...         "minute": [0, 30, 0, 30],  # Minute-level data
    ...         "value": [1, 2, 3, 4],
    ...     }
    ... )
    >>> resampled_data = resample_data_to_hourly(df)
    >>> resampled_data.shape[0]
    # Expecting two rows: one for hour 0 and one for hour 1
    """
    # Create a timestamp from year, month, day, hour, and minute
    if max(data_file["period"]) == 48.0:  # sub-hourly data
        data_file = data_file.with_columns(((pl.col("period") - 1) / 2).cast(pl.Int32).alias("hour"))
    data_file = data_file.with_columns(
        pl.datetime(
            data_file["year"],
            data_file["month"],
            data_file["day"],
            hour=data_file["hour"],
            # minute=data_file["minute"],
        ).alias("timestamp")
    )

    data_file = data_file.drop_nulls().sort("timestamp")

    # Group by the hour and aggregate the values
    return (
        data_file.group_by_dynamic("timestamp", every="1h")
        .agg([pl.col("value").mean().alias("value")])  # Average of values for the hour
        .with_columns(
            pl.col("timestamp").dt.year().alias("year"),
            pl.col("timestamp").dt.month().alias("month"),
            pl.col("timestamp").dt.day().alias("day"),
            pl.col("timestamp").dt.hour().alias("hour"),
            pl.col("value").alias("value"),
        )
        .select(["year", "month", "day", "hour", "value"])
    )


def reconcile_timeseries(data_file: pl.DataFrame, hourly_time_index: pl.DataFrame) -> pl.DataFrame:
    """Adjust timeseries data to match the study year datetime index.

    Parameters
    ----------
    data_file : pl.DataFrame
        The input DataFrame containing the timeseries data to adjust.
    hourly_time_index : pl.DataFrame
        The DataFrame containing the hourly time index for the study year, used for alignment.

    Returns
    -------
    pl.DataFrame
        The adjusted DataFrame after reconciling leap-year or non-leap-year data with the time index.

    Raises
    ------
    AssertionError
        If the provided `hourly_time_index` is empty.

    Notes
    -----
    - If the length of the `data_file` corresponds to a leap year (8784 hours) and the `hourly_time_index` is
      not a leap year, February 29 data is removed.
    - If the length of `data_file` is less than or equal to 8760 hours, missing timestamps are filled to match
      the study year.
    - If the `data_file` contains half-hourly data (17568 or 17520 rows), the data is resampled to an hourly
      frequency.
    """
    assert not hourly_time_index.is_empty()
    leap_year = len(hourly_time_index) == 8784

    if data_file.height in [8784, 8760]:
        if data_file.height == 8784 and not leap_year:
            before_feb_29 = data_file.slice(0, 1416)
            after_feb_29 = data_file.slice(1440, len(data_file) - 1440)
            return pl.concat([before_feb_29, after_feb_29])
        elif data_file.height == 8760 and leap_year:
            return handle_leap_year_adjustment(data_file)
        return data_file

    if data_file.height <= 8760:
        return fill_missing_timestamps(data_file, hourly_time_index)

    if data_file.height in [17568, 17520]:
        return resample_data_to_hourly(data_file)

    return data_file


def construct_pwl_from_quadtratic(fn, mapped_records, num_tranches=6):
    """Given function data of quadratic curve, construct piecewise linear curve with num_tranches tranches."""
    assert isinstance(fn, QuadraticFunctionData), "Input function data must be of type QuadraticFunctionData"
    if isinstance(num_tranches, str):
        num_tranches = int(num_tranches)

    a = fn.quadratic_term
    b = fn.proportional_term
    c = fn.constant_term
    x_min = mapped_records["active_power_limits"].min.magnitude
    x_max = mapped_records["active_power_limits"].max.magnitude

    # Use evenly spaced X values for the tranches
    # Future iteration should accept custom X values for Bid Cost Markup
    x_vals, y_vals = optimize_pwl(a, b, c, x_min, x_max, num_tranches)

    pwl_fn = PiecewiseLinearData(
        points=[XYCoords(x, y) for x, y in sorted(zip(x_vals, y_vals), key=lambda x: x[0])]
    )

    # # Plot the results
    # import matplotlib.pyplot as plt
    # plt.figure(figsize=(8, 6))
    # # Plot the quadratic function
    # x_plot = np.linspace(x_min, x_max, 100)
    # y_quad = a * x_plot**2 + b * x_plot + c
    # plt.plot(x_plot, y_quad, label='Quadratic Function', color='blue', lw=2)
    # plt.plot(x_vals, y_vals, label='Optimized PWL Function', color='red', linestyle='--', marker='o')
    # plt.xlabel('MW')
    # plt.ylabel('MMbtu/hr')
    # plt.title(f'Generator {mapped_records.get("name")} PWL to Quadratic Cost Functions')
    # plt.legend()
    # plt.grid(True)
    # plt.savefig(f'pwl_curves/pwl_optimization_{mapped_records.get("name")}.png')
    # plt.close()

    return pwl_fn


def optimize_pwl(a, b, c, min, max, n_tranches=6):
    y = cp.Variable(n_tranches)
    x_target = np.linspace(min, max, n_tranches)
    y_quad = a * x_target**2 + b * x_target + c
    mse = cp.sum_squares(y - y_quad)

    # Monotonicity constraints
    constraints = [y[i] <= y[i + 1] for i in range(n_tranches - 1)]

    objective = cp.Minimize(mse)
    problem = cp.Problem(objective, constraints)
    problem.solve()

    return x_target, y.value


def _bid_cost_mark_up(fn, mapped_records):
    # TODO(ktehranchi): Implement bid-cost markup
    # First we need to convert whichever type of function we have to a piecewise linear function
    # This PWL function must have X values definted at the mark-up points
    # We can easily modify the mark-up prices by changing the Y values of the PWL function
    # Issue right now is we need to do this for time-varying data but market bid cost isnt implemented
    pass



import pytest
from r2x.config_scenario import Scenario
from r2x.runner import run_single_scenario


@pytest.mark.parametrize(
    "input_model, output_model, name, run_folder",
    [
        ("reeds-US", "sienna", "R2S", "reeds_data_folder"),
        ("reeds-US", "plexos", "R2P", "reeds_data_folder"),
        ("reeds-US", "infrasys", "R2I", "reeds_data_folder"),
        ("infrasys", "sienna", "pjm_2area", "infrasys_data_folder"),
        ("infrasys", "plexos", "pjm_2area", "infrasys_data_folder"),
    ],
    ids=["R2S", "R2P", "R2I", "I2S", "I2P"],
)
def test_full_translation_workflow(input_model, output_model, name, run_folder, request, tmp_folder):
    run_folder = request.getfixturevalue(run_folder)
    config = Scenario.from_kwargs(
        name=name,
        input_model=input_model,
        output_model=output_model,
        run_folder=run_folder,
        output_folder=tmp_folder,
        solve_year=2050,
        model_year=2050,
        weather_year=2012,
    )

    _ = run_single_scenario(scenario=config)



"""Helper functions for upgrader."""

import inspect
from collections.abc import Callable
from r2x.utils import validate_string


def get_function_arguments(argument_input: dict, function: Callable) -> dict:
    """Get arguments to pass to a function based on its signature.

    This function processes the `argument_input` and returns a dictionary of argument
    values that are valid for the given `function`, using the function's signature
    as a filter. String values are validated, nested dictionaries are flattened,
    and only the valid argument keys (as defined in the function signature) are included.

    Parameters
    ----------
    data_dict : dict
        A dictionary containing potential argument values, which may include
        strings, dictionaries, and other types of data.

    function : str
        The name of the function whose signature is used to filter the arguments.

    Returns
    -------
    dict
        A dictionary of filtered arguments that match the function's signature.
        Only arguments that exist in the function's signature will be included.

    Example
    -------
    >>> def example_function(a, b, c=None):
    >>>     pass
    >>> data = {"a": 1, "b": 2, "c": 3, "extra": 4}
    >>> prepare_function_arguments(data, "example_function")
    {'a': 1, 'b': 2, 'c': 3}
    """
    arguments = {}
    for key, value in argument_input.items():
        if isinstance(value, str):
            value = validate_string(value)
        if isinstance(value, dict):
            for sub_key, sub_value in value.items():
                arguments[sub_key] = sub_value
        else:
            arguments[key] = value

    return {key: value for key, value in arguments.items() if key in inspect.getfullargspec(function).args}



"""Compilation of functions that handle upgrades.

This functions apply an update function to certain raw files file before using them for creating the System.
"""

import os
import pathlib
import shutil
import zipfile
from collections import OrderedDict

import pandas as pd
from loguru import logger

from r2x.upgrader.helpers import get_function_arguments
from r2x.utils import read_csv


def rename(fpath: pathlib.Path, new_fname: str) -> pathlib.Path:
    """Apply a new filename to a file.

    This function renames the specified file to a new filename while preserving
    its directory structure. The original file path is resolved to its absolute
    path before renaming.

    Parameters
    ----------
    fpath : pathlib.Path
        The path of the file to be renamed. It should be an existing file.

    new_fname : str
        The new filename to be applied to the file. This should not include any
        directory path; only the name of the file (and optionally the extension).

    Returns
    -------
    pathlib.Path
        The new path of the renamed file.

    Raises
    ------
    FileNotFoundError
        If the file specified by `fpath` does not exist.

    FileExistsError
        If a file with the new name already exists in the same directory.

    Examples
    --------
    >>> import pathlib
    >>> from r2x.upgrader.functions import rename  # Assuming the function is in my_module

    # Rename an existing file
    >>> old_file = pathlib.Path("/path/to/old_file.txt")
    >>> new_file = rename(old_file, "new_file.txt")
    >>> new_file
    PosixPath('/path/to/new_file.txt')

    # Attempt to rename a non-existent file (will raise FileNotFoundError)
    >>> non_existent_file = pathlib.Path("/path/to/non_existent.txt")
    >>> rename(non_existent_file, "new_name.txt")
    FileNotFoundError: [Errno 2] No such file or directory: '/path/to/non_existent.txt'

    # Attempt to rename to a name that already exists (will raise FileExistsError)
    >>> existing_file = pathlib.Path("/path/to/existing_file.txt")
    >>> rename(existing_file, "existing_file.txt")
    FileExistsError: [Errno 17] File exists: '/path/to/existing_file.txt' -> '/path/to/existing_file.txt'
    """
    fpath_new = fpath.resolve().parent.joinpath(new_fname)
    if fpath_new.exists():
        logger.debug(f"File {(fpath_new.name)=} already exist. Skipping it.")
        return fpath_new
    new_fpath = fpath.replace(fpath_new)
    logger.debug("Renaming {} to {}", fpath.name, fpath_new.name)
    return new_fpath


def move_file(fpath: pathlib.Path, new_fpath: str | pathlib.Path) -> pathlib.Path | None:
    """Move a file to a different location.

    This function moves a file from its current location to a new specified
    location using `shutil.move`. If the new path is a string, it will be
    interpreted relative to the parent directory of the original file.

    Parameters
    ----------
    fpath : pathlib.Path
        The path of the file to be moved.
    new_fpath : str or pathlib.Path
        The destination path where the file should be moved. If a string is
        provided, it is interpreted relative to the parent directory of the
        original file.

    Returns
    -------
    pathlib.Path | None
        The new path of the moved file if the move was successful. Returns
        None if the file already exists at the destination.

    Raises
    ------
    FileNotFoundError
        If the specified file does not exist.
    ValueError
        If the new path resolves to a location that already contains a file
        with the same name.

    Examples
    --------
    >>> from r2x.upgrader.functions import move_file
    >>> import pathlib
    >>> move_file(pathlib.Path("/path/to/file.txt"), "new_location/file.txt")

    >>> move_file(pathlib.Path("/path/to/non_existent_file.txt"), "new_location/file.txt")
    Traceback (most recent call last):
        ...
    FileNotFoundError: /path/to/non_existent_file.txt does not exist.

    >>> move_file(pathlib.Path("/path/to/existing_file.txt"), "existing_file.txt")
    DEBUG:__main__:File existing_file.txt already exists in the right place.
    """
    logger.debug(f"Moving {fpath} to {new_fpath}")

    if not fpath.exists():
        raise FileNotFoundError(f"{fpath} does not exist.")

    if not isinstance(new_fpath, pathlib.Path):
        new_fpath = fpath.parent.parent / new_fpath

    if os.path.exists(new_fpath):
        logger.debug(f"File {fpath.name} already exists in the right place.")
        return None

    shutil.move(str(fpath), str(new_fpath))
    return new_fpath


def melt(fpath: pathlib.Path, melt_id_vars: list | None = None, var_name="quarter") -> pd.DataFrame:
    """Apply melt operation to data.

    This function reads a CSV file, performs a melt operation on the specified
    columns, and saves the modified data back to the same file. If no identifier
    variables are provided, default values are used.

    Parameters
    ----------
    fpath : pathlib.Path
        The path to the CSV file to be melted.
    melt_id_vars : list, optional
        A list of columns to use as identifier variables. If None, defaults to
        ["i", "r"].
    melt_var_name : str, optional
        The name to assign to the variable column in the melted DataFrame.
        Default is "quarter".

    Returns
    -------
    pd.DataFrame
        The melted DataFrame

    Raises
    ------
    FileNotFoundError
        If the specified file does not exist.
    ValueError
        If the melt operation fails due to missing identifier variables.

    Examples
    --------
    >>> import pandas as pd
    >>> import tempfile
    >>> import pathlib

    # Create a temporary CSV file for testing
    >>> temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    >>> pd.DataFrame({"i": [1, 2], "r": [3, 4], "Q1": [10, 20], "Q2": [30, 40]}).to_csv(
    ...     temp_file, index=False
    ... )

    >>> melt_season(temp_file)
    >>> pd.read_csv(temp_file)
       i  r quarter  value
    0  1  3      Q1     10
    1  1  3      Q2     30
    2  2  4      Q1     20
    3  2  4      Q2     40

    >>> temp_file.unlink()  # Clean up the temporary file
    """
    if not melt_id_vars:
        melt_id_vars = ["i", "r"]

    if not fpath.exists():
        raise FileNotFoundError(f"{fpath} does not exist.")

    data = pd.read_csv(fpath)

    if var_name in data.columns:
        logger.debug("File {} has been already melted. Skipping.", fpath.name)
        return data

    logger.debug(f"Melting columns for {fpath.name}")
    data = pd.melt(data, id_vars=melt_id_vars, var_name=var_name)

    data.to_csv(fpath, index=False)
    return data


def apply_header(fpath: pathlib.Path, header: str) -> pd.DataFrame | None:
    """Apply a header to a CSV file.

    This function reads the first row of a CSV file and compares it to the
    specified header. If the headers match (case-insensitively), the function
    does nothing. If they don't match, it replaces the existing header with the
    new header and saves the updated data back to the file.

    Parameters
    ----------
    fpath : pathlib.Path
        The path to the CSV file to which the header will be applied.
    header : str
        A comma-separated string representing the new header.

    Returns
    -------
    pd.DataFrame | None
        The modified DataFrame if the header was changed; None if no change was made.

    Raises
    ------
    FileNotFoundError
        If the specified file does not exist.
    ValueError
        If the CSV file is empty or does not have a valid first row.

    Examples
    --------
    >>> import pandas as pd
    >>> import tempfile
    >>> import pathlib

    # Create a temporary CSV file for testing
    >>> temp_file = pathlib.Path(tempfile.gettempdir()) / "test_data.csv"
    >>> pd.DataFrame({"A": [1, 2], "B": [3, 4]}).to_csv(temp_file, index=False, header=None)

    >>> df = apply_header(temp_file, "a,b")
    >>> df
       a  b
    0  1  3
    1  2  4

    >>> df = apply_header(temp_file, "A,B")
    >>> df is None
    True

    >>> temp_file.unlink()  # Clean up the temporary file
    """
    logger.debug(f"Attempting to add header to {fpath.name}.")

    if not fpath.exists():
        raise FileNotFoundError(f"{fpath} does not exist.")

    first_row = pd.read_csv(fpath, header=None, nrows=1).iloc[0].to_list()

    # Lowercase the columns
    first_row = list(map(lambda x: x.lower() if isinstance(x, str) else x, first_row))

    header_row = header.split(",")
    if first_row == header_row:
        logger.debug(f"File {fpath} has appropriate columns")
        return None

    logger.debug(f"Adding columns {header_row} to {fpath.name}")
    data = pd.read_csv(fpath, header=None, names=header_row, skiprows=1)
    data.to_csv(fpath, index=False)
    return data


def set_index(fpath: pathlib.Path, index: str) -> pd.DataFrame | None:
    """Apply a new name for the index of the CSV file at fpath.

    This function reads a CSV file, checks if the index already has a name,
    and if not, assigns a new name to the index and saves the file.

    Parameters
    ----------
    fpath : pathlib.Path
        The path to the CSV file.
    index_rename : str
        The new name to assign to the index.

    Returns
    -------
    pd.DataFrame | None
        The modified DataFrame if the index name was changed, or None if
        the index already had a name or an error occurred.

    Examples
    --------
    >>> import pathlib
    >>> import pandas as pd
    >>> # Assuming a CSV file 'data.csv' with no index name
    >>> index_rename(pathlib.Path("data.csv"), "NewIndexName")
    >>> df = pd.read_csv("data.csv", index_col=0)
    >>> df.index.name
    'NewIndexName'

    >>> # If 'data.csv' already has an index name
    >>> index_rename(pathlib.Path("data.csv"), "AnotherIndexName")
    data.csv already has an index name: 'NewIndexName'
    """
    if not fpath.exists():
        raise FileNotFoundError(f"{fpath} does not exist.")
    data = pd.read_csv(fpath, index_col=0)

    if data.index.name is not None:
        logger.debug(f"{fpath.name} has index name already.")
        return None

    logger.debug(f"Adding index name {index} to {fpath.name}.")
    data.index.name = index
    data.to_csv(fpath)
    return data


def upgrade_handler(run_folder: str | pathlib.Path):
    """Entry point to call the different upgrade functions."""
    logger.info("Starting upgrader")

    # The file tracker has all the information of what update to perform for each data file.
    file_tracker = read_csv("file_tracker.csv", package_data="r2x.upgrader").collect().to_pandas()
    files_to_modify = file_tracker["fname"].unique()

    # This might actually not be safe for the nas.
    f_dict = OrderedDict(
        {
            f.name: f
            for f in pathlib.Path(run_folder).glob("*[inputs_case|outputs]/*")
            if f.name in files_to_modify
        }
    )

    # Backup inputs_case_files for safety
    backup_fpath = pathlib.Path(run_folder).joinpath("backup_files.zip")
    if not backup_fpath.exists():
        logger.info("Creating backup of files.")
        with zipfile.ZipFile(backup_fpath, mode="w") as archive:
            for fname, fpath_name in f_dict.items():
                archive.write(fpath_name, arcname=fname)

    for fname, f_group in file_tracker.groupby("fname", sort=False):
        if fname not in f_dict:
            logger.debug(f"{fname} not in inputs_case_list. Skipping it.")
            continue

        fpath_name = f_dict[fname]
        assert fpath_name
        functions_to_apply = f_group["method"].iloc[0].split(",")  # List of functions
        f_group_dict = f_group.to_dict(orient="records")[
            0
        ]  # Records return a list of dicts. We jsut get the first element
        f_group_dict["fpath"] = fpath_name

        for function in functions_to_apply:
            function_callable = globals()[function]
            function_arguments = get_function_arguments(f_group_dict, function_callable)
            function_callable(**function_arguments)
            # Update f_dict if we renamed a file to apply additional functions to it
            if function == "rename":
                fpath_new = fpath_name.parent.joinpath(f_group_dict["new_fname"][0])
                f_dict[fpath_new.name] = fpath_new



import pytest
from infrasys.time_series_models import SingleTimeSeries

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.models import MonitoredLine, Emission, Generator, PowerLoad
from r2x.parser.handler import get_parser_data
from r2x.parser.reeds import ReEDSParser


@pytest.fixture
def scenario_instance(reeds_data_folder, default_scenario, tmp_folder):
    return Scenario.from_kwargs(
        name=default_scenario,
        input_model="reeds-US",
        output_model="plexos",
        run_folder=reeds_data_folder,
        output_folder=tmp_folder,
        solve_year=2050,
        weather_year=2012,
    )


@pytest.fixture
def reeds_parser_instance(scenario_instance):
    return get_parser_data(scenario_instance, parser_class=ReEDSParser)


def test_reeds_parser_instance(reeds_parser_instance):
    assert isinstance(reeds_parser_instance, ReEDSParser)


def test_parser_has_data(reeds_parser_instance):
    assert len(reeds_parser_instance.data) != 0


def test_system_creation(reeds_parser_instance):
    system = reeds_parser_instance.build_system()
    assert isinstance(system, System)


def test_construct_generators(reeds_parser_instance):
    reeds_parser_instance.system = System(name="Test", auto_add_composed_components=True)
    reeds_parser_instance._construct_buses()
    reeds_parser_instance._construct_reserves()
    reeds_parser_instance._construct_generators()
    generators = [component for component in reeds_parser_instance.system.get_components(Generator)]
    assert all(isinstance(component, Generator) for component in generators)
    assert (
        len([generator for generator in generators]) == 335
    )  # Total number of devices for the pacific scenario for 2050


def test_construct_load_time_series(reeds_parser_instance):
    load_df = reeds_parser_instance.get_data("load").collect()
    reeds_parser_instance.system = System(name="Test")
    reeds_parser_instance._construct_buses()
    reeds_parser_instance._construct_load()
    loads = [component for component in reeds_parser_instance.system.get_components(PowerLoad)]
    assert all(reeds_parser_instance.system.has_time_series(load) for load in loads)

    single_load = loads[0]
    ts: SingleTimeSeries = reeds_parser_instance.system.get_time_series(single_load)
    if len(load_df) > 8760:
        end_idx = 8760 * (reeds_parser_instance.config.weather_year - 2007) + 1  # +1 to be inclusive.
    else:
        end_idx = 8760
    assert len(ts.data) == 8760
    assert len(ts.data) == len(load_df[single_load.bus.name][end_idx - 8760 : end_idx])


@pytest.fixture
def reeds_system(reeds_parser_instance):
    return reeds_parser_instance.build_system()


def test_construct_emissions(reeds_system):
    emission_objects = reeds_system.get_components(Emission)
    emission_objects = [component for component in reeds_system.get_components(Emission)]
    assert all(isinstance(component, Emission) for component in emission_objects)
    assert (
        len(emission_objects) == 136
    )  # Total number of emissions (co2, nox) for all the emitting devices for 2050


def test_construct_branches(reeds_system):
    branch_objects = [component for component in reeds_system.get_components(MonitoredLine)]
    assert all(isinstance(component, MonitoredLine) for component in branch_objects)
    assert len(branch_objects) == 17  # With rating on both direction



from typing import Any
from infrasys.cost_curves import CostCurve, FuelCurve, UnitSystem
from infrasys.function_data import PiecewiseLinearData, QuadraticFunctionData, XYCoords
from infrasys.value_curves import InputOutputCurve, LinearCurve
import pytest

from r2x.config_scenario import Scenario
from r2x.exporter.sienna import SiennaExporter, apply_operation_table_data, get_psy_fields
from r2x.models.costs import ThermalGenerationCost


@pytest.fixture
def scenario_instance(data_folder, tmp_folder):
    return Scenario.from_kwargs(
        name="Test Scenario",
        run_folder=data_folder,
        output_folder=tmp_folder,
        input_model="infrasys",
        output_model="sienna",
        model_year=2010,
    )


@pytest.fixture
def sienna_exporter(scenario_instance, infrasys_test_system, tmp_folder):
    return SiennaExporter(config=scenario_instance, system=infrasys_test_system, output_folder=tmp_folder)


@pytest.mark.sienna
def test_sienna_exporter_instance(sienna_exporter):
    assert isinstance(sienna_exporter, SiennaExporter)


@pytest.mark.sienna
def test_sienna_exporter_run(sienna_exporter, tmp_folder):
    exporter = sienna_exporter.run()

    output_files = [
        "gen.csv",
        "bus.csv",
        "timeseries_pointers.json",
        # "storage.csv",  # Storage is also optional
        "reserves.csv",  # Reserve could be optional
        "dc_branch.csv",
        "branch.csv",
    ]

    for file in output_files:
        assert (tmp_folder / file).exists(), f"File {file} was not created properly."

    # Check that time series was created correctly
    ts_directory = tmp_folder / exporter.ts_directory
    assert any(ts_directory.iterdir())


def test_sienna_exporter_empty_storage(caplog, sienna_exporter):
    sienna_exporter.process_storage_data()
    assert "No storage devices found" in caplog.text


def test_get_psy_fields():
    fields = get_psy_fields()
    assert isinstance(fields, dict)


def test_apply_operation_table_data_basic():
    sample_component = {
        "operation_cost": ThermalGenerationCost(
            variable=FuelCurve(
                value_curve=LinearCurve(10.0, 12),
                vom_cost=LinearCurve(10.0),
                fuel_cost=0.05,
                power_units=UnitSystem.NATURAL_UNITS,
            )
        ).model_dump(mode="python", serialize_as_any=True)
    }

    updated_component = apply_operation_table_data(sample_component)

    assert "variable_cost" in updated_component
    assert "fuel_price" in updated_component
    assert updated_component["variable_cost"] == 10
    assert updated_component["fuel_price"] == 50  # 0.05 * 1000


def test_apply_operation_table_data_heat_rate():
    sample_component = {
        "operation_cost": ThermalGenerationCost(
            variable=FuelCurve(
                value_curve=InputOutputCurve(
                    function_data=QuadraticFunctionData(
                        constant_term=100, proportional_term=20, quadratic_term=0.5
                    )
                ),
                vom_cost=LinearCurve(10.0),
                fuel_cost=0.05,
                power_units=UnitSystem.NATURAL_UNITS,
            )
        ).model_dump(mode="python", serialize_as_any=True)
    }
    updated_component = apply_operation_table_data(sample_component)

    assert "heat_rate_a0" in updated_component
    assert "heat_rate_a1" in updated_component
    assert "heat_rate_a2" in updated_component
    assert updated_component["heat_rate_a0"] == 100
    assert updated_component["heat_rate_a1"] == 20
    assert updated_component["heat_rate_a2"] == 0.5


def test_apply_operation_table_data_cost_curve():
    cost_curve_component = {
        "operation_cost": ThermalGenerationCost(
            variable=CostCurve(
                value_curve=InputOutputCurve(
                    function_data=PiecewiseLinearData(points=[XYCoords(x=0, y=0), XYCoords(x=50, y=1000)])
                ),
                vom_cost=LinearCurve(10.0),
                power_units=UnitSystem.NATURAL_UNITS,
            )
        ).model_dump(mode="python", serialize_as_any=True)
    }
    updated_component = apply_operation_table_data(cost_curve_component)

    assert "output_point_0" in updated_component
    assert "cost_point_0" in updated_component
    assert updated_component["output_point_0"] == 0
    assert updated_component["cost_point_0"] == 0
    assert updated_component["output_point_1"] == 50
    assert updated_component["cost_point_1"] == 1000


def test_apply_operation_table_data_fuel_curve():
    fuel_curve_component = {
        "operation_cost": ThermalGenerationCost(
            variable=FuelCurve(
                value_curve=InputOutputCurve(
                    function_data=PiecewiseLinearData(
                        points=[XYCoords(x=0, y=0), XYCoords(x=50, y=10), XYCoords(x=100, y=25)]
                    )
                ),
                vom_cost=LinearCurve(10.0),
                power_units=UnitSystem.NATURAL_UNITS,
            )
        ).model_dump(mode="python", serialize_as_any=True)
    }
    updated_component = apply_operation_table_data(fuel_curve_component)

    assert "output_point_0" in updated_component
    assert "heat_rate_avg_0" in updated_component
    assert updated_component["output_point_0"] == 0
    assert updated_component["heat_rate_avg_0"] == 0
    assert updated_component["output_point_1"] == 50
    assert updated_component["heat_rate_incr_1"] == 10


def test_apply_operation_table_data_no_operation_cost():
    component = {"id": "test_component"}
    updated_component = apply_operation_table_data(component)
    assert updated_component == component


def test_apply_operation_table_data_no_variable():
    component = {"operation_cost": {}}
    updated_component = apply_operation_table_data(component)
    assert updated_component == component


def test_apply_operation_table_data_unsupported_curve():
    wrong_curve_component: dict[str, Any] = {
        "operation_cost": ThermalGenerationCost(
            variable=FuelCurve(
                value_curve=InputOutputCurve(
                    function_data=PiecewiseLinearData(
                        points=[XYCoords(x=0, y=0), XYCoords(x=50, y=10), XYCoords(x=100, y=25)]
                    )
                ),
                vom_cost=LinearCurve(10.0),
                power_units=UnitSystem.NATURAL_UNITS,
            )
        ).model_dump(mode="python", serialize_as_any=True)
    }
    wrong_curve_component["operation_cost"]["variable_type"] = "NewValueCurve"
    with pytest.raises(NotImplementedError):
        apply_operation_table_data(wrong_curve_component)


def test_apply_operation_table_data_none_fuel_cost():
    component = {"operation_cost": {"variable": {"fuel_cost": None}}}
    with pytest.raises(AssertionError):
        apply_operation_table_data(component)



"""Models for generator devices."""

from typing import Annotated, Any

from pint import Quantity
from pydantic import Field, NonNegativeFloat, field_serializer

from r2x.enums import PrimeMoversType, StorageTechs, ThermalFuels
from r2x.models.core import Device, InputOutput, MinMax, UpDown
from r2x.models.costs import HydroGenerationCost, RenewableGenerationCost, StorageCost, ThermalGenerationCost
from r2x.models.load import PowerLoad
from r2x.models.topology import ACBus
from r2x.units import (
    ActivePower,
    ApparentPower,
    Energy,
    Percentage,
    PowerRate,
    Time,
    VOMPrice,
    ureg,
)


class Generator(Device):
    """Abstract generator class."""

    bus: Annotated[ACBus, Field(description="Bus where the generator is connected.")] | None = None
    rating: Annotated[
        ApparentPower | None,
        Field(ge=0, description="Maximum output power rating of the unit (MVA)."),
    ] = ApparentPower(1, "MVA")
    active_power: Annotated[
        ActivePower,
        Field(
            description=(
                "Initial active power set point of the unit in MW. For power flow, this is the steady "
                "state operating point of the system."
            ),
        ),
    ] = ActivePower(0.0, "MW")
    reactive_power: Annotated[
        ApparentPower | None,
        Field(
            description=(
                "Reactive power set point of the unit in MW. For power flow, this is the steady "
                "state operating point of the system."
            ),
        ),
    ] = ApparentPower(0.0, "MVA")
    base_mva: float = 1
    base_power: Annotated[
        ApparentPower | None,
        Field(
            gt=0,
            description="Base power of the unit (MVA) for per unitization.",
        ),
    ] = None
    must_run: Annotated[int | None, Field(description="If we need to force the dispatch of the device.")] = (
        None
    )
    vom_price: Annotated[VOMPrice, Field(description="Variable operational price $/MWh.")] | None = None
    prime_mover_type: (
        Annotated[PrimeMoversType, Field(description="Prime mover technology according to EIA 923.")] | None
    ) = None
    unit_type: Annotated[
        PrimeMoversType | None, Field(description="Prime mover technology according to EIA 923.")
    ] = None
    min_rated_capacity: Annotated[ActivePower, Field(description="Minimum rated power generation.")] = (
        0 * ureg.MW
    )
    ramp_up: (
        Annotated[
            PowerRate,
            Field(description="Ramping rate on the positve direction."),
        ]
        | None
    ) = None
    ramp_down: (
        Annotated[
            PowerRate,
            Field(description="Ramping rate on the negative direction."),
        ]
        | None
    ) = None
    min_up_time: (
        Annotated[
            Time,
            Field(ge=0, description="Minimum up time in hours for UC decision."),
        ]
        | None
    ) = None
    min_down_time: (
        Annotated[
            Time,
            Field(ge=0, description="Minimum down time in hours for UC decision."),
        ]
        | None
    ) = None
    mean_time_to_repair: (
        Annotated[
            Time,
            Field(gt=0, description="Total hours to repair after outage occur."),
        ]
        | None
    ) = None
    forced_outage_rate: (
        Annotated[
            Percentage,
            Field(description="Expected level of unplanned outages in percent."),
        ]
        | None
    ) = None
    planned_outage_rate: (
        Annotated[
            Percentage,
            Field(description="Expected level of planned outages in percent."),
        ]
        | None
    ) = None
    startup_cost: Annotated[NonNegativeFloat, Field(description="Cost in $ of starting a unit.")] | None = (
        None
    )
    shutdown_cost: (
        Annotated[NonNegativeFloat, Field(description="Cost in $ of shuting down a unit.")] | None
    ) = None
    active_power_limits: Annotated[
        MinMax | None, Field(description="Maximum output power rating of the unit (MVA).")
    ] = None
    reactive_power_limits: Annotated[
        MinMax | None, Field(description="Maximum output power rating of the unit (MVA).")
    ] = None

    @field_serializer("active_power_limits")
    def serialize_active_power_limits(self, min_max: MinMax | dict | None) -> dict[str, Any] | None:
        if min_max is None:
            return None
        if not isinstance(min_max, MinMax):
            min_max = MinMax(**min_max)
        if min_max is not None:
            return {
                "min": min_max.min.magnitude if isinstance(min_max.min, Quantity) else min_max.min,
                "max": min_max.max.magnitude if isinstance(min_max.max, Quantity) else min_max.max,
            }


class RenewableGen(Generator):
    """Abstract class for renewable generators."""


class RenewableDispatch(RenewableGen):
    """Curtailable renewable generator.

    This type of generator have a hourly capacity factor profile.
    """

    power_factor: float = 1.0
    operation_cost: RenewableGenerationCost | None = None


class RenewableNonDispatch(RenewableGen):
    """Non-curtailable renewable generator.

    Renewable technologies w/o operational cost.
    """

    power_factor: float = 1.0


class HydroGen(Generator):
    """Hydroelectric generator."""


class HydroDispatch(HydroGen):
    """Class representing flexible hydro generators."""

    operation_cost: HydroGenerationCost | None = None
    ramp_up: (
        Annotated[
            PowerRate,
            Field(ge=0, description="Ramping rate on the positve direction."),
        ]
        | None
    ) = None
    ramp_down: (
        Annotated[
            PowerRate,
            Field(ge=0, description="Ramping rate on the negative direction."),
        ]
        | None
    ) = None


class HydroFix(HydroGen):
    """Class representing unflexible hydro."""


class HydroEnergyReservoir(HydroGen):
    """Class representing hydro system with reservoirs."""

    ramp_limits: UpDown | None = None
    time_limits: UpDown | None = None
    operation_cost: HydroGenerationCost | None = None
    inflow: float | None = None
    initial_energy: (
        Annotated[NonNegativeFloat, Field(description="Initial water volume or percentage.")] | None
    ) = 0
    initial_storage: float | None = None
    storage_capacity: (
        Annotated[
            Energy,
            Field(description="Total water volume or percentage."),
        ]
        | None
    ) = None
    min_storage_capacity: (
        Annotated[
            Energy,
            Field(description="Minimum water volume or percentage."),
        ]
        | None
    ) = None
    storage_target: (
        Annotated[
            Energy,
            Field(description="Maximum energy limit."),
        ]
        | None
    ) = None


class HydroPumpedStorage(HydroGen):
    """Class representing pumped hydro generators."""

    operation_cost: HydroGenerationCost | StorageCost | None = None
    storage_duration: (
        Annotated[
            Time,
            Field(description="Storage duration in hours."),
        ]
        | None
    ) = None
    initial_volume: (
        Annotated[Energy, Field(gt=0, description="Initial water volume or percentage.")] | None
    ) = None
    storage_capacity: Annotated[
        Energy,
        Field(gt=0, description="Total water volume or percentage."),
    ]
    min_storage_capacity: (
        Annotated[
            Energy,
            Field(description="Minimum water volume or percentage."),
        ]
        | None
    ) = None
    pump_efficiency: Annotated[Percentage, Field(ge=0, le=1, description="Pumping efficiency.")] | None = None
    pump_load: (
        Annotated[
            ActivePower,
            Field(description="Load related to the usage of the pump."),
        ]
        | None
    ) = None

    @classmethod
    def example(cls) -> "HydroPumpedStorage":
        return HydroPumpedStorage(
            name="HydroPumpedStorage",
            active_power=ActivePower(100, "MW"),
            pump_load=ActivePower(100, "MW"),
            bus=ACBus.example(),
            prime_mover_type=PrimeMoversType.PS,
            storage_duration=Time(10, "h"),
            storage_capacity=Energy(1000, "MWh"),
            min_storage_capacity=Energy(10, "MWh"),
            pump_efficiency=Percentage(85, "%"),
            initial_volume=Energy(500, "MWh"),
            ext={"description": "Pumped hydro unit with 10 hour of duration"},
        )


class ThermalGen(Generator):
    """Class representing fuel based thermal generator."""

    fuel: Annotated[str, Field(description="Fuel category")] | None = None
    operation_cost: ThermalGenerationCost | None = None


class ThermalStandard(ThermalGen):
    """Class representing a standard thermal generator."""

    status: bool = True
    ramp_limits: UpDown | None = None
    time_limits: UpDown | None = None
    fuel: ThermalFuels = ThermalFuels.OTHER

    @classmethod
    def example(cls) -> "ThermalStandard":
        return ThermalStandard(
            name="ThermalStandard",
            bus=ACBus.example(),
            fuel=ThermalFuels.NATURAL_GAS,
            active_power=100.0 * ureg.MW,
            ext={"Additional data": "Additional value"},
        )


class ThermalMultiStart(ThermalGen):
    """We will fill this class once we have the need for it."""


class Storage(Generator):
    """Default Storage class."""

    storage_duration: (
        Annotated[
            Time,
            Field(description="Storage duration in hours."),
        ]
        | None
    ) = None
    storage_capacity: Annotated[
        Energy,
        Field(description="Maximum allowed volume or state of charge."),
    ]
    initial_energy: Annotated[Percentage, Field(description="Initial state of charge.")] | None = None
    min_storage_capacity: Annotated[Percentage, Field(description="Minimum state of charge")] = Percentage(
        0, "%"
    )
    max_storage_capacity: Annotated[Percentage, Field(description="Minimum state of charge")] = Percentage(
        100, "%"
    )


class GenericBattery(Storage):
    """Battery energy storage model."""

    charge_efficiency: Annotated[Percentage, Field(ge=0, description="Charge efficiency.")] | None = None
    discharge_efficiency: Annotated[Percentage, Field(ge=0, description="Discharge efficiency.")] | None = (
        None
    )


class EnergyReservoirStorage(Storage):
    """Battery energy storage model."""

    storage_technology_type: StorageTechs
    operation_cost: StorageCost | None = None
    charge_efficiency: Annotated[Percentage, Field(ge=0, description="Charge efficiency.")] | None = None
    discharge_efficiency: Annotated[Percentage, Field(ge=0, description="Discharge efficiency.")] | None = (
        None
    )
    storage_level_limits: MinMax | None = None
    initial_storage_capacity_level: float
    input_active_power_limits: MinMax
    output_active_power_limits: MinMax
    efficiency: InputOutput


class HybridSystem(Device):
    """Representation of hybrid system with renewable generation, load, thermal generation and storage.

    This class is just a link between two components.
    For the implementation see:
    - https://github.com/NREL-Sienna/PowerSystems.jl/blob/main/src/models/HybridSystem.jl
    """

    storage_unit: Storage | None = None
    renewable_unit: RenewableGen | None = None
    thermal_unit: ThermalGen | None = None
    electric_load: PowerLoad | None = None
